@@ -1,5 +1,9 @@
+from __future__ import annotations
+
+
 __copyright__ = """
 Copyright (C) 2009-2013 Andreas Kloeckner
+Copyright (C) 2013- University of Illinois Board of Trustees
 Copyright (C) 2020 Matt Wala
 """
 
@@ -25,49 +29,31 @@ THE SOFTWARE.
 
 import builtins
 import logging
-import math
 import operator
 import re
 import sys
+from collections.abc import (
+    Callable,
+    Collection,
+    Hashable,
+    Iterable,
+    Iterator,
+    Mapping,
+    Sequence,
+)
 from functools import reduce, wraps
 from sys import intern
 from typing import (
-    TYPE_CHECKING,
     Any,
-    Callable,
     ClassVar,
-    Collection,
-    Dict,
+    Concatenate,
     Generic,
-    Hashable,
-    Iterable,
-    List,
-    Mapping,
-    Optional,
-    Sequence,
-    Set,
-    Tuple,
-    Type,
+    ParamSpec,
+    Protocol,
     TypeVar,
-    Union,
 )
 
-
-if TYPE_CHECKING:
-    # NOTE: mypy seems to be confused by the `try.. except` below when called with
-    #   python -m mypy --python-version 3.8 ...
-    # see https://github.com/python/mypy/issues/14220
-    from typing_extensions import Concatenate, ParamSpec, SupportsIndex
-else:
-    try:
-        from typing import Concatenate, SupportsIndex
-    except ImportError:
-        from typing_extensions import Concatenate, SupportsIndex
-
-    try:
-        from typing import ParamSpec
-    except ImportError:
-        from typing_extensions import ParamSpec  # type: ignore[assignment]
+from typing_extensions import Self, dataclass_transform
 
 
 # These are deprecated and will go away in 2022.
@@ -83,8 +69,6 @@ Math
 ----
 
 .. autofunction:: levi_civita
-.. autofunction:: perm
-.. autofunction:: comb
 
 Assertive accessors
 -------------------
@@ -127,6 +111,7 @@ Permutations, Tuples, Integer sequences
 .. autofunction:: generate_all_integer_tuples_below
 .. autofunction:: generate_permutations
 .. autofunction:: generate_unique_permutations
+.. autoclass:: _ConcatenableSequence
 
 Formatting
 ----------
@@ -158,6 +143,7 @@ Deprecation Warnings
 --------------------
 
 .. autofunction:: deprecate_keyword
+.. autofunction:: module_getattr_for_deprecations
 
 Functions for dealing with (large) auxiliary files
 --------------------------------------------------
@@ -221,6 +207,11 @@ Python's built-in set type, they maintain the internal order of elements.
 .. autofunction:: unique_intersection
 .. autofunction:: unique_union
 
+Functionality for dataclasses
+-----------------------------
+
+.. autofunction:: opt_frozen_dataclass
+
 Type Variables Used
 -------------------
 
@@ -250,8 +241,10 @@ P = ParamSpec("P")
 
 # {{{ code maintenance
 
+# Undocumented on purpose for now, unclear that this is a great idea, given
+# that typing.deprecated exists.
 class MovedFunctionDeprecationWrapper:
-    def __init__(self, f: F, deadline: Optional[Union[int, str]] = None) -> None:
+    def __init__(self, f: F, deadline: int | str | None = None) -> None:
         if deadline is None:
             deadline = "the future"
 
@@ -268,8 +261,8 @@ class MovedFunctionDeprecationWrapper:
 
 
 def deprecate_keyword(oldkey: str,
-        newkey: Optional[str] = None, *,
-        deadline: Optional[str] = None):
+        newkey: str | None = None, *,
+        deadline: str | None = None):
     """Decorator used to deprecate function keyword arguments.
 
     :arg oldkey: deprecated argument name.
@@ -308,6 +301,36 @@ def deprecate_keyword(oldkey: str,
 
     return wrapper
 
+
+def module_getattr_for_deprecations(
+            module_name: str,
+            depr_name_to_replacement_and_obj: Mapping[
+                str, tuple[str, object, str | int]
+            ],
+            name: str
+        ) -> object:
+    """A helper to construct module-level :meth:`object.__getattr__` functions
+    so that deprecated names can still be found but raise a warning.
+
+    The typical usage pattern is as follows::
+
+        __getattr__ = partial(module_getattr_for_deprecations, __name__, {
+            "OldName": ("NewName", NewName, 2026),
+            })
+    """
+
+    replacement_and_obj = depr_name_to_replacement_and_obj.get(name, None)
+    if replacement_and_obj is not None:
+        replacement, obj, deadline = replacement_and_obj
+        from warnings import warn
+
+        warn(f"'{module_name}.{name}' is deprecated. "
+                f"Use '{replacement}' instead. "
+                f"'{module_name}.{name}' will continue to work until {deadline}.",
+                DeprecationWarning, stacklevel=2)
+        return obj
+    raise AttributeError(name)
+
 # }}}
 
 
@@ -316,11 +339,10 @@ def deprecate_keyword(oldkey: str,
 def delta(x, y):
     if x == y:
         return 1
-    else:
-        return 0
+    return 0
 
 
-def levi_civita(tup: Tuple[int, ...]) -> int:
+def levi_civita(tup: tuple[int, ...]) -> int:
     """Compute an entry of the Levi-Civita symbol for the indices *tuple*."""
     if len(tup) == 2:
         i, j = tup
@@ -328,77 +350,7 @@ def levi_civita(tup: Tuple[int, ...]) -> int:
     if len(tup) == 3:
         i, j, k = tup
         return (j-i) * (k-i) * (k-j) // 2
-    else:
-        raise NotImplementedError(f"Levi-Civita symbol in {len(tup)} dimensions")
-
-
-factorial = MovedFunctionDeprecationWrapper(math.factorial, deadline=2023)
-
-try:
-    # NOTE: only available in python >= 3.8
-    perm = MovedFunctionDeprecationWrapper(math.perm, deadline=2023)
-except AttributeError:
-    def _unchecked_perm(n, k):
-        result = 1
-        while k:
-            result *= n
-            n -= 1
-            k -= 1
-
-        return result
-
-    def perm(n: SupportsIndex,              # type: ignore[misc]
-             k: Optional[SupportsIndex] = None) -> int:
-        """
-        :returns: :math:`P(n, k)`, the number of permutations of length :math:`k`
-            drawn from :math:`n` choices.
-        """
-        from warnings import warn
-        warn("This function is deprecated and will go away in 2023. "
-                "Use `math.perm` instead, which is available from Python 3.8.",
-                DeprecationWarning, stacklevel=2)
-
-        if k is None:
-            return math.factorial(n)
-
-        import operator
-        n, k = operator.index(n), operator.index(k)
-        if k > n:
-            return 0
-
-        if k < 0:
-            raise ValueError("k must be a non-negative integer")
-
-        if n < 0:
-            raise ValueError("n must be a non-negative integer")
-
-        from numbers import Integral
-        if not isinstance(k, Integral):
-            raise TypeError(f"'{type(k).__name__}' object cannot be interpreted "
-                            "as an integer")
-
-        if not isinstance(n, Integral):
-            raise TypeError(f"'{type(n).__name__}' object cannot be interpreted "
-                            "as an integer")
-
-        return _unchecked_perm(n, k)
-
-try:
-    # NOTE: only available in python >= 3.8
-    comb = MovedFunctionDeprecationWrapper(math.comb, deadline=2023)
-except AttributeError:
-    def comb(n: SupportsIndex,              # type: ignore[misc]
-             k: SupportsIndex) -> int:
-        """
-        :returns: :math:`C(n, k)`, the number of combinations (subsets)
-            of length :math:`k` drawn from :math:`n` choices.
-        """
-        from warnings import warn
-        warn("This function is deprecated and will go away in 2023. "
-                "Use `math.comb` instead, which is available from Python 3.8.",
-                DeprecationWarning, stacklevel=2)
-
-        return _unchecked_perm(n, k) // math.factorial(k)
+    raise NotImplementedError(f"Levi-Civita symbol in {len(tup)} dimensions")
 
 
 def norm_1(iterable):
@@ -436,12 +388,12 @@ class RecordWithoutPickling:
     will be individually derived from this class.
     """
 
-    __slots__: ClassVar[List[str]] = []
-    fields: ClassVar[Set[str]]
+    __slots__: ClassVar[list[str]] = []
+    fields: ClassVar[set[str]]
 
     def __init__(self,
-                 valuedict: Optional[Mapping[str, Any]] = None,
-                 exclude: Optional[Sequence[str]] = None,
+                 valuedict: Mapping[str, Any] | None = None,
+                 exclude: Sequence[str] | None = None,
                  **kwargs: Any) -> None:
         assert self.__class__ is not Record
 
@@ -497,7 +449,7 @@ class RecordWithoutPickling:
 
 
 class Record(RecordWithoutPickling):
-    __slots__: ClassVar[List[str]] = []
+    __slots__: ClassVar[list[str]] = []
 
     def __getstate__(self):
         return {
@@ -713,7 +665,7 @@ def memoize(*args: F, **kwargs: Any) -> F:
 
     use_kw = bool(kwargs.pop("use_kwargs", False))
 
-    default_key_func: Optional[Callable[..., Any]]
+    default_key_func: Callable[..., Any] | None
 
     if use_kw:
         def default_key_func(*inner_args, **inner_kwargs):
@@ -784,7 +736,7 @@ class _HasKwargs:
 
 def memoize_on_first_arg(
         function: Callable[Concatenate[T, P], R], *,
-        cache_dict_name: Optional[str] = None) -> Callable[Concatenate[T, P], R]:
+        cache_dict_name: str | None = None) -> Callable[Concatenate[T, P], R]:
     """Like :func:`memoize_method`, but for functions that take the object
     in which do memoization information is stored as first argument.
 
@@ -814,9 +766,8 @@ def memoize_on_first_arg(
         if attribute_error:
             object.__setattr__(obj, cache_dict_name, {key: result})
             return result
-        else:
-            getattr(obj, cache_dict_name)[key] = result
-            return result
+        getattr(obj, cache_dict_name)[key] = result
+        return result
 
     def clear_cache(obj):
         object.__delattr__(obj, cache_dict_name)
@@ -862,7 +813,7 @@ class keyed_memoize_on_first_arg(Generic[T, P, R]):  # noqa: N801
 
     def __init__(self,
             key: Callable[P, Hashable], *,
-            cache_dict_name: Optional[str] = None) -> None:
+            cache_dict_name: str | None = None) -> None:
         self.key = key
         self.cache_dict_name = cache_dict_name
 
@@ -966,14 +917,14 @@ class memoize_in:  # noqa: N801
             try:
                 return self.cache_dict[args]
             except KeyError:
-                result = inner(*args)
+                result = inner(*args, **kwargs)
                 self.cache_dict[args] = result
                 return result
 
         return new_inner
 
 
-class keyed_memoize_in(Generic[P, R]):  # noqa: N801
+class keyed_memoize_in(Generic[P]):  # noqa: N801
     """Like :class:`memoize_in`, but additionally uses a function *key* to
     compute the key under which the function result is memoized.
 
@@ -1000,12 +951,12 @@ class keyed_memoize_in(Generic[P, R]):  # noqa: N801
         @wraps(inner)
         def new_inner(*args: P.args, **kwargs: P.kwargs) -> R:
             assert not kwargs
-            key = self.key(*args)
+            key = self.key(*args, **kwargs)
 
             try:
                 return self.cache_dict[key]
             except KeyError:
-                result = inner(*args)
+                result = inner(*args, **kwargs)
                 self.cache_dict[key] = result
                 return result
 
@@ -1059,7 +1010,7 @@ def monkeypatch_class(_name, bases, namespace):
 # {{{ generic utilities
 
 def add_tuples(t1, t2):
-    return tuple(t1v + t2v for t1v, t2v in zip(t1, t2))
+    return tuple(t1v + t2v for t1v, t2v in zip(t1, t2, strict=True))
 
 
 def negate_tuple(t1):
@@ -1107,7 +1058,7 @@ def general_sum(sequence):
 
 def linear_combination(coefficients, vectors):
     result = coefficients[0] * vectors[0]
-    for c, v in zip(coefficients[1:], vectors[1:]):
+    for c, v in zip(coefficients[1:], vectors[1:], strict=True):
         result += c*v
     return result
 
@@ -1277,8 +1228,7 @@ def argmin2(iterable, return_value=False):
 
     if return_value:
         return current_argmin, current_min
-    else:
-        return current_argmin
+    return current_argmin
 
 
 def argmax2(iterable, return_value=False):
@@ -1295,8 +1245,7 @@ def argmax2(iterable, return_value=False):
 
     if return_value:
         return current_argmax, current_max
-    else:
-        return current_argmax
+    return current_argmax
 
 
 def argmin(iterable):
@@ -1384,13 +1333,10 @@ class VarianceAggregator:
         if self.entire_pop:
             if self.n == 0:
                 return None
-            else:
-                return self.m2/self.n
-        else:
-            if self.n <= 1:
-                return None
-            else:
-                return self.m2/(self.n - 1)
+            return self.m2/self.n
+        if self.n <= 1:
+            return None
+        return self.m2/(self.n - 1)
 
 
 def variance(iterable, entire_pop):
@@ -1411,7 +1357,9 @@ def std_deviation(iterable, finite_pop):
 
 # {{{ permutations, tuples, integer sequences
 
-def wandering_element(length, wanderer=1, landscape=0):
+def wandering_element(length: int,
+                      wanderer: int = 1,
+                      landscape: int = 0) -> Iterator[tuple[int, ...]]:
     for i in range(length):
         yield i*(landscape,) + (wanderer,) + (length-1-i)*(landscape,)
 
@@ -1427,18 +1375,21 @@ def indices_in_shape(shape):
     if not shape:
         yield ()
     elif len(shape) == 1:
-        for i in range(0, shape[0]):
+        for i in range(shape[0]):
             yield (i,)
     else:
         remainder = shape[1:]
-        for i in range(0, shape[0]):
+        for i in range(shape[0]):
             for rest in indices_in_shape(remainder):
                 yield (i, *rest)
 
 
-def generate_nonnegative_integer_tuples_below(n, length=None, least=0):
+def generate_nonnegative_integer_tuples_below(
+        n: Sequence[int] | int, length: int | None = None, least: int = 0
+        ) -> Iterator[tuple[int, ...]]:
     """n may be a sequence, in which case length must be None."""
     if length is None:
+        assert not isinstance(n, int)
         if not n:
             yield ()
             return
@@ -1447,6 +1398,7 @@ def generate_nonnegative_integer_tuples_below(n, length=None, least=0):
         n = n[1:]
         next_length = None
     else:
+        assert isinstance(n, int)
         my_n = n
 
         assert length >= 0
@@ -1463,12 +1415,12 @@ def generate_nonnegative_integer_tuples_below(n, length=None, least=0):
 
 
 def generate_decreasing_nonnegative_tuples_summing_to(
-        n, length, min_value=0, max_value=None):
+        n: int, length: int, min_value: int = 0, max_value: int | None = None
+        ) -> Iterator[tuple[int, ...]]:
     if length == 0:
         yield ()
     elif length == 1:
-        if n <= max_value:
-            # print "MX", n, max_value
+        if max_value is None or n <= max_value:
             yield (n,)
         else:
             return
@@ -1476,14 +1428,14 @@ def generate_decreasing_nonnegative_tuples_summing_to(
         if max_value is None or n < max_value:
             max_value = n
 
-        for i in range(min_value, max_value+1):
-            # print "SIG", sig, i
+        for i in range(min_value, max_value + 1):
             for remainder in generate_decreasing_nonnegative_tuples_summing_to(
-                    n-i, length-1, min_value, i):
+                    n - i, length - 1, min_value=min_value, max_value=i):
                 yield (i, *remainder)
 
 
-def generate_nonnegative_integer_tuples_summing_to_at_most(n, length):
+def generate_nonnegative_integer_tuples_summing_to_at_most(
+        n: int, length: int) -> Iterator[tuple[int, ...]]:
     """Enumerate all non-negative integer tuples summing to at most n,
     exhausting the search space by varying the first entry fastest,
     and the last entry the slowest.
@@ -1502,24 +1454,56 @@ def generate_nonnegative_integer_tuples_summing_to_at_most(n, length):
 generate_positive_integer_tuples_below = generate_nonnegative_integer_tuples_below
 
 
-def _pos_and_neg_adaptor(tuple_iter):
+def _pos_and_neg_adaptor(
+        tuple_iter: Iterator[tuple[int, ...]]
+    ) -> Iterator[tuple[int, ...]]:
     for tup in tuple_iter:
         nonzero_indices = [i for i in range(len(tup)) if tup[i] != 0]
         for do_neg_tup in generate_nonnegative_integer_tuples_below(
                 2, len(nonzero_indices)):
+
             this_result = list(tup)
             for index, do_neg in enumerate(do_neg_tup):
                 if do_neg:
                     this_result[nonzero_indices[index]] *= -1
+
             yield tuple(this_result)
 
 
-def generate_all_integer_tuples_below(n, length, least_abs=0):
+def generate_all_integer_tuples_below(
+        n: int, length: int, least_abs: int = 0
+    ) -> Iterator[tuple[int, ...]]:
     return _pos_and_neg_adaptor(generate_nonnegative_integer_tuples_below(
         n, length, least_abs))
 
 
-def generate_permutations(original):
+T_co = TypeVar("T_co", covariant=True)
+
+
+class _ConcatenableSequence(Generic[T_co], Protocol):
+    """
+    A protocol that supports the following:
+
+    .. automethod:: __getitem__
+    .. automethod:: __add__
+    .. automethod:: __len__
+    """
+    def __getitem__(self, slice) -> Self:
+        ...
+
+    def __add__(self, other: Self) -> Self:
+        ...
+
+    def __len__(self) -> int:
+        ...
+
+    def __iter__(self) -> Iterator[T_co]:
+        ...
+
+
+def generate_permutations(
+            original: _ConcatenableSequence[T]
+        ) -> Iterator[_ConcatenableSequence[T]]:
     """Generate all permutations of the list *original*.
 
     Nicked from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/252178
@@ -1529,12 +1513,17 @@ def generate_permutations(original):
     else:
         for perm_ in generate_permutations(original[1:]):
             for i in range(len(perm_)+1):
-                # nb str[0:1] works in both string and list contexts
+                # NOTE: ary[0:1] works in both string and list contexts
                 yield perm_[:i] + original[0:1] + perm_[i:]
 
 
-def generate_unique_permutations(original):
+def generate_unique_permutations(
+            original: _ConcatenableSequence[T]
+        ) -> Iterator[_ConcatenableSequence[T]]:
     """Generate all unique permutations of the list *original*.
+
+    Note that, unlike for :func:`generate_permutations`, *original* must be a
+    hashable object.
     """
 
     had_those = set()
@@ -1569,26 +1558,30 @@ a_star = MovedFunctionDeprecationWrapper(a_star_moved)
 class Table:
     """An ASCII table generator.
 
-    .. attribute:: nrows
-    .. attribute:: ncolumns
-
-    .. attribute:: alignments
-
-        A :class:`tuple` of alignments of each column: ``"l"``, ``"c"``, or ``"r"``,
-        for left, center, and right alignment, respectively). Columns which
-        have no alignment specifier will use the last specified alignment. For
-        example, with ``alignments=("l", "r")``, the third and all following
-        columns will use right alignment.
-
+    .. automethod:: __init__
     .. automethod:: add_row
 
+    .. autoproperty:: nrows
+    .. autoproperty:: ncolumns
+
     .. automethod:: __str__
     .. automethod:: github_markdown
     .. automethod:: csv
     .. automethod:: latex
+    .. automethod:: text_without_markup
     """
 
-    def __init__(self, alignments: Optional[Tuple[str, ...]] = None) -> None:
+    def __init__(self, alignments: tuple[str, ...] | None = None) -> None:
+        """Create a new :class:`Table`.
+
+        :arg alignments: A :class:`tuple` of alignments of each column:
+            ``"l"``, ``"c"``, or ``"r"``, for left, center, and right
+            alignment, respectively). Columns which have no alignment specifier
+            will use the last specified alignment. For example, with
+            ``alignments=("l", "r")``, the third and all following
+            columns will use right alignment.
+        """
+
         if alignments is None:
             alignments = ("l",)
         else:
@@ -1597,18 +1590,22 @@ class Table:
 
             alignments = tuple(alignments)
 
-        self.rows: List[Tuple[str, ...]] = []
+        self.rows: list[tuple[str, ...]] = []
         self.alignments = alignments
 
     @property
     def nrows(self) -> int:
+        """The number of rows currently in the table."""
         return len(self.rows)
 
     @property
     def ncolumns(self) -> int:
+        """The number of columns currently in the table."""
         return len(self.rows[0])
 
-    def add_row(self, row: Tuple[Any, ...]) -> None:
+    def add_row(self, row: tuple[Any, ...]) -> None:
+        """Add *row* to the table. Note that all rows must have the same number
+        of columns."""
         if self.rows and len(row) != self.ncolumns:
             raise ValueError(
                     f"tried to add a row with {len(row)} columns to "
@@ -1616,14 +1613,15 @@ class Table:
 
         self.rows.append(tuple(str(i) for i in row))
 
-    def _get_alignments(self) -> Tuple[str, ...]:
+    def _get_alignments(self) -> tuple[str, ...]:
         # NOTE: If not all alignments were specified, extend alignments with the
         # last alignment specified
-        return (self.alignments
+        return (
+                self.alignments
                 + (self.alignments[-1],) * (self.ncolumns - len(self.alignments))
-                )
+                )[:self.ncolumns]
 
-    def _get_column_widths(self, rows) -> Tuple[int, ...]:
+    def _get_column_widths(self, rows) -> tuple[int, ...]:
         return tuple(
             max(len(row[i]) for row in rows) for i in range(self.ncolumns)
             )
@@ -1650,13 +1648,13 @@ class Table:
         col_widths = self._get_column_widths(self.rows)
 
         lines = [" | ".join([
-            cell.center(col_width) if align == "c"
-            else cell.ljust(col_width) if align == "l"
-            else cell.rjust(col_width)
-            for cell, col_width, align in zip(row, col_widths, alignments)])
+            cell.center(cwidth) if align == "c"
+            else cell.ljust(cwidth) if align == "l"
+            else cell.rjust(cwidth)
+            for cell, cwidth, align in zip(row, col_widths, alignments, strict=True)])
             for row in self.rows]
-        lines[1:1] = ["+".join("-" * (col_width + 1 + (i > 0))
-            for i, col_width in enumerate(col_widths))]
+        lines[1:1] = ["+".join("-" * (cwidth + 1 + (i > 0))
+            for i, cwidth in enumerate(col_widths))]
 
         return "\n".join(lines)
 
@@ -1688,22 +1686,23 @@ class Table:
         col_widths = self._get_column_widths(rows)
 
         lines = [" | ".join([
-            cell.center(col_width) if align == "c"
-            else cell.ljust(col_width) if align == "l"
-            else cell.rjust(col_width)
-            for cell, col_width, align in zip(row, col_widths, alignments)])
+            cell.center(cwidth) if align == "c"
+            else cell.ljust(cwidth) if align == "l"
+            else cell.rjust(cwidth)
+            for cell, cwidth, align in zip(row, col_widths, alignments, strict=True)])
             for row in rows]
         lines[1:1] = ["|".join(
-            (":" + "-" * (col_width - 1 + (i > 0)) + ":") if align == "c"
-            else (":" + "-" * (col_width + (i > 0))) if align == "l"
-            else ("-" * (col_width + (i > 0)) + ":")
-            for i, (col_width, align) in enumerate(zip(col_widths, alignments)))]
+            (":" + "-" * (cwidth - 1 + (i > 0)) + ":") if align == "c"
+            else (":" + "-" * (cwidth + (i > 0))) if align == "l"
+            else ("-" * (cwidth + (i > 0)) + ":")
+            for i, (cwidth, align) in enumerate(
+                zip(col_widths, alignments, strict=True)))]
 
         return "\n".join(lines)
 
     def csv(self,
             dialect: str = "excel",
-            csv_kwargs: Optional[Dict[str, Any]] = None) -> str:
+            csv_kwargs: dict[str, Any] | None = None) -> str:
         """Returns a string containing a CSV representation of the table.
 
         :arg dialect: String passed to :func:`csv.writer`.
@@ -1740,7 +1739,7 @@ class Table:
 
     def latex(self,
             skip_lines: int = 0,
-            hline_after: Optional[Tuple[int, ...]] = None) -> str:
+            hline_after: tuple[int, ...] | None = None) -> str:
         r"""Returns a string containing the rows of a LaTeX representation of
         the table.
 
@@ -1772,9 +1771,41 @@ class Table:
 
         return "\n".join(lines)
 
+    def text_without_markup(self) -> str:
+        """Returns a string representation of the table without markup.
+
+        .. doctest::
+
+            >>> tbl = Table()
+            >>> tbl.add_row([0, "orange"])
+            >>> tbl.add_row([1111, "apple"])
+            >>> tbl.add_row([2, "pear"])
+            >>> print(tbl.text_without_markup())
+            0    orange
+            1111 apple
+            2    pear
+        """
+        if not self.rows:
+            return ""
+
+        alignments = self._get_alignments()
+        col_widths = self._get_column_widths(self.rows)
+
+        lines = [" ".join([
+            cell.center(cwidth) if align == "c"
+            else cell.ljust(cwidth) if align == "l"
+            else cell.rjust(cwidth)
+            for cell, cwidth, align in zip(row, col_widths, alignments, strict=True)])
+            for row in self.rows]
+
+        # Remove the extra space added by the last cell
+        lines = [line.rstrip() for line in lines]
+
+        return "\n".join(lines)
+
 
 def merge_tables(*tables: Table,
-        skip_columns: Optional[Tuple[int, ...]] = None) -> Table:
+        skip_columns: tuple[int, ...] | None = None) -> Table:
     """
     :arg skip_columns: a :class:`tuple` of column indices to skip in all the
         tables except the first one.
@@ -1792,10 +1823,9 @@ def merge_tables(*tables: Table,
     def remove_columns(i, row):
         if i == 0 or skip_columns is None:
             return row
-        else:
-            return tuple(
-                entry for i, entry in enumerate(row) if i not in skip_columns
-                )
+        return tuple(
+            entry for i, entry in enumerate(row) if i not in skip_columns
+            )
 
     alignments = sum((
         remove_columns(i, tbl._get_alignments())
@@ -1848,15 +1878,14 @@ def string_histogram(
     if use_unicode:
         def format_bar(cnt):
             scaled = cnt*width/max_count
-            full = int(floor(scaled))
-            eighths = int(ceil((scaled-full)*8))
+            full = floor(scaled)
+            eighths = ceil((scaled-full)*8)
             if eighths:
                 return full*chr(0x2588) + chr(0x2588+(8-eighths))
-            else:
-                return full*chr(0x2588)
+            return full*chr(0x2588)
     else:
         def format_bar(cnt):
-            return int(ceil(cnt*width/max_count))*"#"
+            return ceil(cnt*width/max_count)*"#"
 
     max_count = max(bins)
     total_count = sum(bins)
@@ -1865,7 +1894,7 @@ def string_histogram(
         bin_value,
         bin_value/total_count*100,
         format_bar(bin_value))
-        for bin_start, bin_value in zip(bin_starts, bins))
+        for bin_start, bin_value in zip(bin_starts, bins, strict=True))
 
 # }}}
 
@@ -1987,7 +2016,7 @@ class StderrToStdout:
 
 
 def typedump(val: Any, max_seq: int = 5,
-             special_handlers: Optional[Mapping[Type, Callable]] = None,
+             special_handlers: Mapping[type, Callable] | None = None,
              fully_qualified_name: bool = True) -> str:
     """
     Return a string representation of the type of *val*, recursing into
@@ -2018,13 +2047,11 @@ def typedump(val: Any, max_seq: int = 5,
         if type(obj).__module__ == "builtins":
             if fully_qualified_name:
                 return type(obj).__qualname__
-            else:
-                return type(obj).__name__
+            return type(obj).__name__
 
         if fully_qualified_name:
             return type(obj).__module__ + "." + type(obj).__qualname__
-        else:
-            return type(obj).__name__
+        return type(obj).__name__
 
     # Special handling for 'str' since it is also iterable
     if isinstance(val, str):
@@ -2046,10 +2073,9 @@ def typedump(val: Any, max_seq: int = 5,
                 t = ",".join(typedump(x, max_seq, special_handlers)
                             for x in val[:max_seq])
                 return f"{objname(val)}({t},...)"
-            else:
-                t = ",".join(typedump(x, max_seq, special_handlers)
-                            for x in val)
-                return f"{objname(val)}({t})"
+            t = ",".join(typedump(x, max_seq, special_handlers)
+                        for x in val)
+            return f"{objname(val)}({t})"
 
         except TypeError:
             return objname(val)
@@ -2189,12 +2215,10 @@ def common_dtype(dtypes, default=None):
     dtypes = list(dtypes)
     if dtypes:
         return argmax2((dtype, dtype.num) for dtype in dtypes)
-    else:
-        if default is not None:
-            return default
-        else:
-            raise ValueError(
-                    "cannot find common dtype of empty dtype list")
+    if default is not None:
+        return default
+    raise ValueError(
+            "cannot find common dtype of empty dtype list")
 
 
 def to_uncomplex_dtype(dtype):
@@ -2212,13 +2236,10 @@ def match_precision(dtype, dtype_to_match):
     if dtype_is_complex:
         if tgt_is_double:
             return numpy.dtype(numpy.complex128)
-        else:
-            return numpy.dtype(numpy.complex64)
-    else:
-        if tgt_is_double:
-            return numpy.dtype(numpy.float64)
-        else:
-            return numpy.dtype(numpy.float32)
+        return numpy.dtype(numpy.complex64)
+    if tgt_is_double:
+        return numpy.dtype(numpy.float64)
+    return numpy.dtype(numpy.float32)
 
 # }}}
 
@@ -2238,7 +2259,7 @@ UNIQUE_NAME_GEN_COUNTER_RE = re.compile(r"^(?P<based_on>\w+)_(?P<counter>\d+)$")
 
 
 def generate_numbered_unique_names(
-        prefix: str, num: Optional[int] = None) -> Iterable[Tuple[int, str]]:
+        prefix: str, num: int | None = None) -> Iterable[tuple[int, str]]:
     if num is None:
         yield (0, prefix)
         num = 0
@@ -2265,7 +2286,7 @@ class UniqueNameGenerator:
     .. automethod:: __call__
     """
     def __init__(self,
-            existing_names: Optional[Collection[str]] = None,
+            existing_names: Collection[str] | None = None,
             forced_prefix: str = ""):
         """
         Create a new :class:`UniqueNameGenerator`.
@@ -2279,7 +2300,7 @@ class UniqueNameGenerator:
 
         self.existing_names = set(existing_names)
         self.forced_prefix = forced_prefix
-        self.prefix_to_counter: Dict[str, int] = {}
+        self.prefix_to_counter: dict[str, int] = {}
 
     def is_name_conflicting(self, name: str) -> bool:
         """Returns *True* if *name* conflicts with an existing :class:`str`."""
@@ -2293,7 +2314,6 @@ class UniqueNameGenerator:
             This will not get called for the names in the *existing_names*
             argument to :meth:`__init__`.
         """
-        pass
 
     def add_name(self, name: str, *, conflicting_ok: bool = False) -> None:
         """
@@ -2597,12 +2617,11 @@ class ProcessLogger:
         if sys.stdin is None:
             # Can happen, e.g., if pudb is controlling the console.
             use_late_start_logging = False
+        elif hasattr(sys.stdin, "closed") and not sys.stdin.closed:
+            # can query stdin.isatty() only if stdin's open
+            use_late_start_logging = sys.stdin.isatty()
         else:
-            if hasattr(sys.stdin, "closed") and not sys.stdin.closed:
-                # can query stdin.isatty() only if stdin's open
-                use_late_start_logging = sys.stdin.isatty()
-            else:
-                use_late_start_logging = False
+            use_late_start_logging = False
 
         import os
         if os.environ.get("PYTOOLS_LOG_NO_THREADS", ""):
@@ -2749,44 +2768,14 @@ def resolve_name(name):
 
     .. versionadded:: 2021.1.2
     """
-    # Delete the tail of the function and deprecate this once we require Python 3.9.
-    if sys.version_info >= (3, 9):
-        # use the official version
-        import pkgutil
-        return pkgutil.resolve_name(name)
+    from warnings import warn
 
-    import importlib
+    warn("'pytools.resolve_name' is deprecated and will be removed in 2024. "
+         "Use 'pkgutil.resolve_name' from the standard library instead.",
+         DeprecationWarning, stacklevel=2)
 
-    m = _NAME_PATTERN.match(name)
-    if not m:
-        raise ValueError(f"invalid format: {name!r}")
-    groups = m.groups()
-    if groups[2]:
-        # there is a colon - a one-step import is all that's needed
-        mod = importlib.import_module(groups[0])
-        parts = groups[3].split(".") if groups[3] else []
-    else:
-        # no colon - have to iterate to find the package boundary
-        parts = name.split(".")
-        modname = parts.pop(0)
-        # first part *must* be a module/package.
-        mod = importlib.import_module(modname)
-        while parts:
-            p = parts[0]
-            s = f"{modname}.{p}"
-            try:
-                mod = importlib.import_module(s)
-                parts.pop(0)
-                modname = s
-            except ImportError:
-                break
-    # if we reach this point, mod is the module, already imported, and
-    # parts is the list of parts in the object hierarchy to be traversed, or
-    # an empty list if just the module is wanted.
-    result = mod
-    for p in parts:
-        result = getattr(result, p)
-    return result
+    import pkgutil
+    return pkgutil.resolve_name(name)
 
 # }}}
 
@@ -2795,7 +2784,7 @@ def resolve_name(name):
 
 def unordered_hash(hash_instance: Any,
                    iterable: Iterable[Any],
-                   hash_constructor: Optional[Callable[[], Any]] = None) -> Any:
+                   hash_constructor: Callable[[], Any] | None = None) -> Any:
     """Using a hash algorithm given by the parameter-less constructor
     *hash_constructor*, return a hash object whose internal state
     depends on the entries of *iterable*, but not their order. If *hash*
@@ -2851,7 +2840,7 @@ def sphere_sample_equidistant(npoints_approx: int, r: float = 1.0):
     """
 
     import numpy as np
-    points: List[np.ndarray] = []
+    points: list[np.ndarray] = []
 
     count = 0
     a = 4 * np.pi / npoints_approx
@@ -2891,7 +2880,7 @@ _SPHERE_FIBONACCI_OFFSET = (
 
 def sphere_sample_fibonacci(
         npoints: int, r: float = 1.0, *,
-        optimize: Optional[str] = None):
+        optimize: str | None = None):
     """Generate points on a sphere based on an offset Fibonacci lattice from [2]_.
 
     .. [2] http://extremelearning.com.au/how-to-evenly-distribute-points-on-a-sphere-more-effectively-than-the-canonical-fibonacci-lattice/
@@ -2931,7 +2920,7 @@ def sphere_sample_fibonacci(
 
 # {{{ strtobool
 
-def strtobool(val: Optional[str], default: Optional[bool] = None) -> bool:
+def strtobool(val: str | None, default: bool | None = None) -> bool:
     """Convert a string representation of truth to True or False.
     True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
     are 'n', 'no', 'f', 'false', 'off', and '0'.  Uppercase versions are
@@ -2956,13 +2945,12 @@ def strtobool(val: Optional[str], default: Optional[bool] = None) -> bool:
     val = val.lower()
     if val in ("y", "yes", "t", "true", "on", "1"):
         return True
-    elif val in ("n", "no", "f", "false", "off", "0"):
+    if val in ("n", "no", "f", "false", "off", "0"):
         return False
-    else:
-        raise ValueError(f"invalid truth value '{val}'. "
-                          "Valid values are ('y', 'yes', 't', 'true', 'on', '1') "
-                          "for 'True' and ('n', 'no', 'f', 'false', 'off', '0') "
-                          "for 'False'. Uppercase versions are also accepted.")
+    raise ValueError(f"invalid truth value '{val}'. "
+                      "Valid values are ('y', 'yes', 't', 'true', 'on', '1') "
+                      "for 'True' and ('n', 'no', 'f', 'false', 'off', '0') "
+                      "for 'False'. Uppercase versions are also accepted.")
 
 # }}}
 
@@ -3038,7 +3026,7 @@ def unique_union(*args: Iterable[T]) -> Collection[T]:
     if not args:
         return []
 
-    res: Dict[T, None] = {}
+    res: dict[T, None] = {}
     for seq in args:
         for item in seq:
             if item not in res:
@@ -3049,6 +3037,76 @@ def unique_union(*args: Iterable[T]) -> Collection[T]:
 # }}}
 
 
+@dataclass_transform(frozen_default=True)
+def opt_frozen_dataclass(
+            *,
+            init: bool = True,
+            repr: bool = True,
+            eq: bool = True,
+            order: bool = False,
+            unsafe_hash: bool | None = None,
+            match_args: bool = True,
+            kw_only: bool = False,
+            slots: bool = False,
+            # Added in 3.11.
+            weakref_slot: bool = False,
+         ) -> Callable[[type[T]], type[T]]:
+    """Like :func:`dataclasses.dataclass`, but marks the dataclass frozen
+    only if :data:`__debug__` is active. Frozen dataclasses have a ~20%
+    cost penalty (on creation, from having to call :meth:`object.__setattr__`) that
+    this decorator avoids when the interpreter runs with "optimization"
+    enabled.
+
+    The resulting dataclass supports hashing, even when it is not actually frozen,
+    if *unsafe_hash* is left at the default or set to *True*.
+
+    .. note::
+
+        Python prevents non-frozen dataclasses from inheriting from frozen ones,
+        and vice versa. To ensure frozen-ness is applied predictably in all
+        scenarios (mainly :data:`__debug__` on and off), it is strongly recommended
+        that all dataclasses inheriting from ones with this decorator *also*
+        use this decorator. There are no run-time checks to make sure of this.
+
+    .. versionadded:: 2024.1.18
+    """
+    def map_cls(cls: type[T]) -> type[T]:
+        # This ensures that the resulting dataclass is hashable with and without
+        # __debug__, unless the user overrides unsafe_hash or provides their own
+        # __hash__ method.
+        if unsafe_hash is None:
+            if (eq
+                    and not __debug__
+                    and "__hash__" not in cls.__dict__):
+                loc_unsafe_hash = True
+            else:
+                loc_unsafe_hash = False
+        else:
+            loc_unsafe_hash = unsafe_hash
+
+        dc_extra_kwargs: dict[str, bool] = {}
+        if weakref_slot:
+            if sys.version_info < (3, 11):
+                raise TypeError("weakref_slot is not available before Python 3.11")
+            dc_extra_kwargs["weakref_slot"] = weakref_slot
+
+        from dataclasses import dataclass
+        return dataclass(
+             init=init,
+             repr=repr,
+             eq=eq,
+             order=order,
+             unsafe_hash=loc_unsafe_hash,
+             frozen=__debug__,
+             match_args=match_args,
+             kw_only=kw_only,
+             slots=slots,
+             **dc_extra_kwargs,
+        )(cls)
+
+    return map_cls
+
+
 def _test():
     import doctest
     doctest.testmod()
