@@ -5,9 +5,9 @@
 .. autoclass:: PConvergenceVerifier
 """
 
+from __future__ import annotations
 
 import numbers
-from typing import List, Optional, Tuple
 
 import numpy as np
 
@@ -45,7 +45,7 @@ class EOCRecorder:
     """
 
     def __init__(self) -> None:
-        self.history: List[Tuple[float, float]] = []
+        self.history: list[tuple[float, float]] = []
 
     def add_data_point(self, abscissa: float, error: float) -> None:
         if not (isinstance(abscissa, numbers.Number)
@@ -60,7 +60,7 @@ class EOCRecorder:
         self.history.append((abscissa, error))
 
     def estimate_order_of_convergence(self,
-            gliding_mean: Optional[int] = None,
+            gliding_mean: int | None = None,
             ) -> np.ndarray:
         abscissae = np.array([a for a, e in self.history])
         errors = np.array([e for a, e in self.history])
@@ -82,8 +82,7 @@ class EOCRecorder:
         return result
 
     def order_estimate(self) -> float:
-        from typing import cast
-        return cast(float, self.estimate_order_of_convergence()[0, 1])
+        return self.estimate_order_of_convergence()[0, 1]
 
     def max_error(self) -> float:
         return max(err for absc, err in self.history)
@@ -134,14 +133,13 @@ class EOCRecorder:
 
         if table_type == "markdown":
             return tbl.github_markdown()
-        elif table_type == "latex":
+        if table_type == "latex":
             return tbl.latex()
-        elif table_type == "ascii":
+        if table_type == "ascii":
             return str(tbl)
-        elif table_type == "csv":
+        if table_type == "csv":
             return tbl.csv()
-        else:
-            raise ValueError(f"unknown table type: {table_type}")
+        raise ValueError(f"unknown table type: {table_type}")
 
     def __str__(self):
         return self.pretty_print()
@@ -159,7 +157,7 @@ class EOCRecorder:
 
 
 def stringify_eocs(*eocs: EOCRecorder,
-        names: Optional[Tuple[str, ...]] = None,
+        names: tuple[str, ...] | None = None,
         abscissa_label: str = "h",
         error_label: str = "Error",
         gliding_mean: int = 2,
@@ -186,19 +184,18 @@ def stringify_eocs(*eocs: EOCRecorder,
         error_format=error_format,
         eoc_format=eoc_format,
         gliding_mean=gliding_mean)
-        for name, eoc in zip(names, eocs)
+        for name, eoc in zip(names, eocs, strict=True)
         ], skip_columns=(0,))
 
     if table_type == "markdown":
         return tbl.github_markdown()
-    elif table_type == "latex":
+    if table_type == "latex":
         return tbl.latex()
-    elif table_type == "ascii":
+    if table_type == "ascii":
         return str(tbl)
-    elif table_type == "csv":
+    if table_type == "csv":
         return tbl.csv()
-    else:
-        raise ValueError(f"unknown table type: {table_type}")
+    raise ValueError(f"unknown table type: {table_type}")
 
 # }}}
 
@@ -219,7 +216,7 @@ class PConvergenceVerifier:
         tbl = Table()
         tbl.add_row(("p", "error"))
 
-        for p, err in zip(self.orders, self.errors):
+        for p, err in zip(self.orders, self.errors, strict=True):
             tbl.add_row((str(p), str(err)))
 
         return str(tbl)
