@@ -28,7 +28,6 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 """
 
-
 __doc__ = """
 Graph Algorithms
 ================
@@ -54,6 +53,10 @@ Graph Algorithms
 Type Variables Used
 -------------------
 
+.. class:: _SupportsLT
+
+    A :class:`~typing.Protocol` for `__lt__` support.
+
 .. class:: NodeT
 
     Type of a graph node, can be any hashable type.
@@ -67,41 +70,25 @@ Type Variables Used
     is included as a key in the graph.
 """
 
-from dataclasses import dataclass
-from typing import (
-    TYPE_CHECKING,
-    Any,
+from collections.abc import (
     Callable,
     Collection,
-    Dict,
-    FrozenSet,
-    Generic,
     Hashable,
     Iterable,
     Iterator,
-    List,
     Mapping,
     MutableSet,
-    Optional,
+)
+from dataclasses import dataclass
+from typing import (
+    Any,
+    Generic,
     Protocol,
-    Set,
-    Tuple,
+    TypeAlias,
     TypeVar,
 )
 
 
-if TYPE_CHECKING:
-    # NOTE: mypy seems to be confused by the `try.. except` below when called with
-    #   python -m mypy --python-version 3.8 ...
-    # see https://github.com/python/mypy/issues/14220
-    from typing_extensions import TypeAlias
-else:
-    try:
-        from typing import TypeAlias
-    except ImportError:
-        from typing_extensions import TypeAlias
-
-
 NodeT = TypeVar("NodeT", bound=Hashable)
 
 GraphT: TypeAlias[NodeT] = Mapping[NodeT, Collection[NodeT]]
@@ -115,7 +102,7 @@ def reverse_graph(graph: GraphT[NodeT]) -> GraphT[NodeT]:
 
     :returns: A :class:`dict` representing *graph* with edges reversed.
     """
-    result: Dict[NodeT, Set[NodeT]] = {}
+    result: dict[NodeT, set[NodeT]] = {}
 
     for node_key, successor_nodes in graph.items():
         # Make sure every node is in the result even if it has no successors
@@ -131,11 +118,18 @@ def reverse_graph(graph: GraphT[NodeT]) -> GraphT[NodeT]:
 
 # {{{ a_star
 
+@dataclass(frozen=True)
+class _AStarNode(Generic[NodeT]):
+    state: NodeT
+    parent: _AStarNode[NodeT] | None
+    path_cost: float | int
+
+
 def a_star(
         initial_state: NodeT, goal_state: NodeT, neighbor_map: GraphT[NodeT],
-        estimate_remaining_cost: Optional[Callable[[NodeT], float]] = None,
+        estimate_remaining_cost: Callable[[NodeT], float] | None = None,
         get_step_cost: Callable[[Any, NodeT], float] = lambda x, y: 1
-        ) -> List[NodeT]:
+        ) -> list[NodeT]:
     """
     With the default cost and heuristic, this amounts to Dijkstra's algorithm.
     """
@@ -146,22 +140,13 @@ def a_star(
         def estimate_remaining_cost(x: NodeT) -> float:
             if x != goal_state:
                 return 1
-            else:
-                return 0
-
-    class AStarNode:
-        __slots__ = ["parent", "path_cost", "state"]
-
-        def __init__(self, state: NodeT, parent: Any, path_cost: float) -> None:
-            self.state = state
-            self.parent = parent
-            self.path_cost = path_cost
+            return 0
 
     inf = float("inf")
     init_remcost = estimate_remaining_cost(initial_state)
     assert init_remcost != inf
 
-    queue = [(init_remcost, AStarNode(initial_state, parent=None, path_cost=0))]
+    queue = [(init_remcost, _AStarNode(initial_state, parent=None, path_cost=0))]
     visited_states = set()
 
     while queue:
@@ -170,7 +155,7 @@ def a_star(
 
         if top.state == goal_state:
             result = []
-            it: Optional[AStarNode] = top
+            it: _AStarNode[NodeT] | None = top
             while it is not None:
                 result.append(it.state)
                 it = it.parent
@@ -188,7 +173,7 @@ def a_star(
             estimated_path_cost = top.path_cost+step_cost+remaining_cost
             heappush(queue,
                 (estimated_path_cost,
-                    AStarNode(state, top, path_cost=top.path_cost + step_cost)))
+                    _AStarNode(state, top, path_cost=top.path_cost + step_cost)))
 
     raise RuntimeError("no solution")
 
@@ -197,21 +182,20 @@ def a_star(
 
 # {{{ compute SCCs with Tarjan's algorithm
 
-def compute_sccs(graph: GraphT[NodeT]) -> List[List[NodeT]]:
+def compute_sccs(graph: GraphT[NodeT]) -> list[list[NodeT]]:
     to_search = set(graph.keys())
-    visit_order: Dict[NodeT, int] = {}
+    visit_order: dict[NodeT, int] = {}
     scc_root = {}
     sccs = []
 
     while to_search:
         top = next(iter(to_search))
-        call_stack: List[Tuple[NodeT, Iterator[NodeT], Optional[NodeT]]] = [(top,
-                                                                 iter(graph[top]),
-                                                                 None)]
+        call_stack: list[tuple[NodeT, Iterator[NodeT], NodeT | None]] = (
+            [(top, iter(graph[top]), None)])
         visit_stack = []
         visiting = set()
 
-        scc: List[NodeT] = []
+        scc: list[NodeT] = []
 
         while call_stack:
             top, children, last_popped_child = call_stack.pop()
@@ -269,7 +253,7 @@ class CycleError(Exception):
 
 
 class _SupportsLT(Protocol):
-    def __lt__(self, other: object) -> bool:
+    def __lt__(self, other: Any) -> bool:
         ...
 
 
@@ -289,9 +273,10 @@ class _HeapEntry(Generic[NodeT]):
         return self.key < other.key
 
 
-def compute_topological_order(graph: GraphT[NodeT],
-                              key: Optional[Callable[[NodeT], Any]] = None) \
-                              -> List[NodeT]:
+def compute_topological_order(
+        graph: GraphT[NodeT],
+        key: Callable[[NodeT], _SupportsLT] | None = None,
+        ) -> list[NodeT]:
     """Compute a topological order of nodes in a directed graph.
 
     :arg key: A custom key function may be supplied to determine the order in
@@ -415,8 +400,8 @@ def contains_cycle(graph: GraphT[NodeT]) -> bool:
 
 # {{{ compute induced subgraph
 
-def compute_induced_subgraph(graph: Mapping[NodeT, Set[NodeT]],
-                             subgraph_nodes: Set[NodeT]) -> GraphT[NodeT]:
+def compute_induced_subgraph(graph: Mapping[NodeT, set[NodeT]],
+                             subgraph_nodes: set[NodeT]) -> GraphT[NodeT]:
     """Compute the induced subgraph formed by a subset of the vertices in a
     graph.
 
@@ -446,9 +431,9 @@ def compute_induced_subgraph(graph: Mapping[NodeT, Set[NodeT]],
 # {{{ as_graphviz_dot
 
 def as_graphviz_dot(graph: GraphT[NodeT],
-                    node_labels: Optional[Callable[[NodeT], str]] = None,
-                    edge_labels: Optional[Callable[[NodeT, NodeT], str]] = None) \
-                    -> str:
+                    node_labels: Callable[[NodeT], str] | None = None,
+                    edge_labels: Callable[[NodeT, NodeT], str] | None = None,
+                    ) -> str:
     """
     Create a visualization of the graph *graph* in the
     `dot <http://graphviz.org/>`__ language.
@@ -486,7 +471,7 @@ def as_graphviz_dot(graph: GraphT[NodeT],
     # Add nodes
     content = "\n".join(
         [f'{node_to_id[node]} [label="{dot_escape(node_labels(node))}"];'
-         for node in node_to_id.keys()])
+         for node in node_to_id])
 
     content += "\n"
 
@@ -509,7 +494,7 @@ def validate_graph(graph: GraphT[NodeT]) -> None:
     Validates that all successor nodes of each node in *graph* are keys in
     *graph* itself. Raises a :class:`ValueError` if not.
     """
-    seen_nodes: Set[NodeT] = set()
+    seen_nodes: set[NodeT] = set()
 
     for children in graph.values():
         seen_nodes.update(children)
@@ -556,7 +541,7 @@ def is_connected(graph: GraphT[NodeT]) -> bool:
 
 
 def undirected_graph_from_edges(
-            edges: Iterable[Tuple[NodeT, NodeT]],
+            edges: Iterable[tuple[NodeT, NodeT]],
         ) -> GraphT[NodeT]:
     """
     Constructs an undirected graph using *edges*.
@@ -565,7 +550,7 @@ def undirected_graph_from_edges(
 
     :returns: A :class:`GraphT` that is the undirected graph.
     """
-    undirected_graph: Dict[NodeT, Set[NodeT]] = {}
+    undirected_graph: dict[NodeT, set[NodeT]] = {}
 
     for lhs, rhs in edges:
         if lhs == rhs:
@@ -580,24 +565,42 @@ def undirected_graph_from_edges(
 
 def get_reachable_nodes(
         undirected_graph: GraphT[NodeT],
-        source_node: NodeT) -> FrozenSet[NodeT]:
+        source_node: NodeT,
+        exclude_nodes: Collection[NodeT] | None = None) -> frozenset[NodeT]:
     """
     Returns a :class:`frozenset` of all nodes in *undirected_graph* that are
     reachable from *source_node*.
+
+    If any node from *exclude_nodes* lies on a path between *source_node* and
+    some other node :math:`u` in *undirected_graph* and there are no other
+    viable paths, then :math:`u` is considered not reachable from *source_node*.
+
+    In the case where *source_node* is in *exclude_nodes*, then no node is
+    reachable from *source_node*, so an empty :class:`frozenset` is returned.
     """
-    nodes_visited: Set[NodeT] = set()
+    if exclude_nodes is not None and source_node in exclude_nodes:
+        return frozenset()
+
+    nodes_visited: set[NodeT] = set()
+    reachable_nodes: set[NodeT] = set()
     nodes_to_visit = {source_node}
 
+    if exclude_nodes is None:
+        exclude_nodes = set()
+
     while nodes_to_visit:
         current_node = nodes_to_visit.pop()
         nodes_visited.add(current_node)
 
-        neighbors = undirected_graph[current_node]
-        nodes_to_visit.update({node
-                               for node in neighbors
-                               if node not in nodes_visited})
+        reachable_nodes.add(current_node)
 
-    return frozenset(nodes_visited)
+        neighbors = undirected_graph[current_node]
+        nodes_to_visit.update({
+            node for node in neighbors
+            if node not in nodes_visited and node not in exclude_nodes
+        })
+
+    return frozenset(reachable_nodes)
 
 
 # vim: foldmethod=marker
